/*
 * signal-event.c - signal handling
 *
 * Grigoriy Sitkarev, <sitkarev@unixkomi.ru>
 */

#include <sys/signal.h>
#include <sys/fcntl.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

#include "common.h"
#include "macros.h"
#include "xmalloc.h"
#include "failures.h"
#include "signal-event.h"

struct sigevt {
	struct sigevt		*next;
	int 			signo;
	sigeventcallback_t	callback;
	void			*opaque;
	int			no_delay;
};

static int pipe_pair[2] = { -1, -1 };
static struct sigevt *sigevent_table[NSIG];

int sigevent_add(int signo, sigeventcallback_t callback, void *opaque, int no_delay)
{
	struct sigevt *evt;

	if (!callback)
		return -1;
	if (signo <= 0 || signo >= NSIG)
		return -1;

	evt = xmalloc0(sizeof(*evt));
	evt->signo = signo;
	evt->no_delay = no_delay;
	evt->callback = callback;
	evt->opaque = opaque;
	/* I think the following is atomic.. */
	evt->next = sigevent_table[signo];
	sigevent_table[signo] = evt;
	return 0;
}

int sigevent_del(int signo, sigeventcallback_t callback, void *opaque)
{
	int res;
	struct sigevt *evt, *prev = NULL;
	struct sigaction act;

	if (!callback)
		return -1;
	if (signo <= 0 || signo >= NSIG)
		return -1;

	evt = sigevent_table[signo];
	while (evt) {
		if (evt->callback == callback &&
		    evt->opaque == opaque) {
			if (prev)
				prev->next = evt->next;
			else
				sigevent_table[signo] = evt->next;
			xfree(evt);
			/* Set signal back to it's defaul when no handlers left */
			if (sigevent_table[signo] == NULL) {
				memset(&act, 0, sizeof(struct sigaction));
				act.sa_flags = 0;
				sigfillset(&act.sa_mask);
				act.sa_handler = SIG_DFL;
				res = sigaction(signo, &act, NULL);
				if (res != 0)
					xwarning("sigevent_del: signo=%d sigaction()=%d: %s",
							signo, res, strerror(errno));
			}
			return 0;
		}
		prev = evt;
	}
	return -1;
}

int sigevent_ignore(int signo)
{
	int res;
	struct sigaction act;

	if (signo <= 0 || signo >= NSIG)
		return -1;

	memset(&act, 0, sizeof(act));
	act.sa_flags = 0;
	sigfillset(&act.sa_mask);
	act.sa_handler = SIG_IGN;
	res = sigaction(signo, &act, NULL);
	if (res == -1)
		xwarning("sigevent_ignore: sigaction() can't set signo=%d", signo);
	return 0;
}

static void signal_handler(int signo, siginfo_t *si, void *opaque)
{
	int with_delay = 0;
	int saved_errno;
	unsigned char blah;
	struct sigevt *evt;

	/* Shouldn't happen but who knows the kernel ?? :) */
	if (signo <= 0 || signo >= NSIG)
		return;

	saved_errno = errno;
	for (evt = sigevent_table[signo]; evt != NULL; evt = evt->next) {
		if (evt->no_delay)
			(*evt->callback)(signo, si, evt->opaque);
		else
			with_delay++;
	}
	if (with_delay) {
		blah = signo;
		write(pipe_pair[1], &blah, sizeof(blah));
	}
	errno = saved_errno;
}

void sigevent_run()
{
	int i, res, signo;
	unsigned char sigs[32];
	unsigned char already_run[NSIG];
	struct sigevt *evt;

	do {
		res = read(pipe_pair[0], &sigs, sizeof(sigs));
		if (res <= 0)
			break;
		memset(already_run, 0, sizeof(already_run));
		for (i = 0; i < res; i++) {
			signo = sigs[i];
			if (signo <= 0 || signo >= NSIG) {
				xwarning("sigevent_run: got signo=%d??", signo);
				continue;
			}
			if (already_run[signo])
				continue;
			for (evt = sigevent_table[signo]; evt != NULL; evt = evt->next) {
				if (evt->no_delay)
					continue;
				(*evt->callback)(signo, NULL, evt->opaque);
			}
			already_run[signo] = 1;
		}
	} while (res > 0);
}

int sigevent_fd()
{
	return pipe_pair[0];
}

int sigevent_init()
{
	int i;
	int res;
	int flags;
	struct sigaction act;

	if (pipe_pair[0] >= 0 || pipe_pair[1] >= 0)
		xerror(1, "sigevent_init: called twice!");

	res = pipe(pipe_pair);
	if (res == -1)
		xerror(1, "sigevent_init: pipe() failed: %s", strerror(errno));
	flags = fcntl(pipe_pair[0], F_GETFL);
	res = fcntl(pipe_pair[0], F_SETFL, flags | O_NONBLOCK);
	if (res == -1)
		xerror(1, "sigevent_init: fcntl() can't set O_NONBLOCK on fd=%d: %s",
				pipe_pair[0], strerror(errno));
	flags = fcntl(pipe_pair[1], F_GETFL);
	res = fcntl(pipe_pair[1], F_SETFL, flags | O_NONBLOCK);
	if (res == -1)
		xerror(1, "sigevent_init: fcntl() can't set O_NONBLOCK on fd=%d: %s",
				pipe_pair[1], strerror(errno));

	for (i = 1; i < NSIG; i++) {
		if (sigevent_table[i] != NULL) {
			memset(&act, 0, sizeof(act));
			act.sa_flags = SA_SIGINFO;
			sigfillset(&act.sa_mask);
			act.sa_sigaction = signal_handler;
			res = sigaction(i, &act, NULL);
			if (res == -1)
				xwarning("sigevent_init: sigaction() failed to set "
						"handler signo=%d: %s",
						i, strerror(errno));
		}
	}
	return 0;
}

void sigevent_deinit()
{
	int i;
	struct sigevt *evt, *next;

	if (pipe_pair[0] >= 0) {
		close(pipe_pair[0]);
		pipe_pair[0] = -1;
	}
	if (pipe_pair[1] >= 0) {
		close(pipe_pair[1]);
		pipe_pair[1] = -1;
	}
	for (i = 1; i < NSIG; i++) {
		/* I believe this is atomic */
		evt = sigevent_table[i];
		sigevent_table[i] = NULL;
		while (evt != NULL) {
			next = evt->next;
			xfree(evt);
			evt = next;
		}
	}
}
